from pathlib import Path

from cuesplitter.models import Album, Track

from mutagen.flac import FLAC


def set_tags(track_path: Path, album: Album, track: Track) -> None:
    """add vorbis commets"""
    f = FLAC(track_path)

    f.clear()
    f.clear_pictures()

    # Album info
    if album.title:
        f['ALBUM'] = album.title
    if album.performer:
        f['ARTIST'] = album.performer
        f['PERFORMER'] = album.performer
    if album.rem.date:
        f['DATE'] = str(album.rem.date)
    if album.rem.genre:
        f['GENRE'] = album.rem.genre
    if album.rem.replaygain_gain:
        f['REPLAYGAIN_ALBUM_GAIN'] = f'{album.rem.replaygain_gain:.2f} dB'
    if album.rem.replaygain_peak:
        f['REPLAYGAIN_ALBUM_PEAK'] = f'{album.rem.replaygain_peak:.6f}'

    # Track info
    if track.performer:
        f['ARTIST'] = track.performer
        f['PERFORMER'] = track.performer
    if track.title:
        f['TITLE'] = track.title
    f['TRACKNUMBER'] = f'{track.track:02d}'
    if track.rem.replaygain_gain:
        f['REPLAYGAIN_TRACK_GAIN'] = f'{track.rem.replaygain_gain:.2f} dB'
    if track.rem.replaygain_peak:
        f['REPLAYGAIN_TRACK_PEAK'] = f'{track.rem.replaygain_peak:.6f}'

    f.save()



from pathlib import Path
from cuesplitter.core import split_album, verify_album

from cuetools import CueParseError, CueValidationError

import typer

from rich.console import Console

import asyncio

import time


app = typer.Typer()

stdout = Console()
stderr = Console(stderr=True)


@app.command()
def split(
    input: Path,
    output: Path = Path(),
    strict: bool = False,
    dry: bool = False,
    verify: bool = False,
    workers: int = 1,
    timer: bool = False,
):
    """
    Split album on different tracks by `.cue` file
    """
    t1 = time.time()

    try:
        result = asyncio.run(split_album(input, output, strict, workers, dry, verify))

        out = sorted(result)
        for path in out:
            stdout.print(path, end='\0')
    except CueValidationError as e:
        stderr.print('[bold red]Cue validation error:[/bold red]')
        stderr.print(str(e))
        raise typer.Exit(code=1)
    except CueParseError as e:
        stderr.print('[bold red]Cue parse error:[/bold red]')
        stderr.print(str(e))
        raise typer.Exit(code=1)
    except RuntimeError as e:
        stderr.print('[bold red]Runtime error:[/bold red]')
        stderr.print(str(e))
        raise typer.Exit(code=1)
    if timer:
        t2 = time.time()
        stdout.print(f'[bold green]{(t2 - t1)}[/bold green]')


# @app.command()
# def join(
#     tracks: list[Path],
#     output: Path = Path('joined.flac'),
# ):
#     """
#     Join tracks  into a single album file
#     """
#     if not tracks:
#         stderr.print('Error: No input tracks provided.')
#         raise typer.Exit(1)

#     for p in tracks:
#         if not p.exists():
#             stderr.print(f'Error: File not found: {p}')
#             raise typer.Exit(1)

#     join_album(tracks, output)


@app.command()
def verify(original: Path, tracks: list[Path], workers: int = 1):
    """Verify album split"""
    if not tracks:
        stderr.print('Error: No input tracks provided.')
        raise typer.Exit(1)

    for p in tracks:
        if not p.exists():
            stderr.print(f'Error: File not found: {p}')
            raise typer.Exit(1)

    stdout.print(asyncio.run(verify_album(tracks, original, workers)))


if __name__ == '__main__':
    app()



import asyncio
from pathlib import Path

from typing import Any, Awaitable, Callable
import cuetools

from cuesplitter.models import Album, Track

from cuesplitter.tags import set_tags

import tempfile

from cuesplitter.ffmpeg import extract_track, join_tracks, get_raw_pcm, cmp_raw_pcm


async def parse_album(cue_path: Path, strict_title_case: bool) -> Album:
    cue_dir = cue_path.parent

    with open(cue_path, 'r') as cue:
        album = await Album.from_album_data(
            cuetools.load(cue, strict_title_case), cue_dir
        )

    return album


async def execute_by_workers(
    queue: list[tuple[Any, ...]],
    handler: Callable[..., Awaitable[Any]],
    num_workers: int,
) -> list[Any]:
    """Executes all tasks in the queue using the specified number of workers. Workers take data from the queue as needed and process it with the specified handlers until the queue is empty."""

    async_queue: asyncio.Queue[tuple[Any] | None] = asyncio.Queue()

    for item in queue:
        await async_queue.put(item)

    output: list[Any] = []

    workers = [
        asyncio.create_task(worker(async_queue, output, handler))
        for _ in range(num_workers)
    ]

    await async_queue.join()
    for _ in range(num_workers):
        await async_queue.put(None)

    await asyncio.gather(*workers)

    return output


async def worker(
    queue: asyncio.Queue[tuple[Any, ...] | None],
    output: list[Any],
    handler: Callable[..., Awaitable[Any]],
) -> None:
    """Coroutine wrapper that allows it to work with an asynchronous queue"""
    while True:
        item = await queue.get()
        try:
            if not item:
                break

            result = await handler(*item)
            output.append(result)
        finally:
            queue.task_done()


async def track_extraction_handler(
    track: Track,
    album: Album,
    output_dir: Path,
    dry: bool,
) -> Path:
    file_name = f'{track.track:02d}'
    if track.title:
        file_name += f' - {track.title.replace("'", "")}.flac'

    output_file = output_dir / file_name

    if not dry:
        await extract_track(track.offset, track.duration, track.file, output_file)
        set_tags(output_file, album, track)

    return (output_file).resolve()


async def split_album(
    cue_path: Path,
    output_dir: Path,
    strict_title_case: bool,
    num_workers: int,
    dry: bool,
    verify: bool,
) -> list[Path]:
    album = await parse_album(cue_path, strict_title_case)

    if not dry:
        output_dir.mkdir(parents=True, exist_ok=True)

    queue = [(track, album, output_dir, dry) for track in album.tracks]

    output_paths: list[Path] = await execute_by_workers(
        queue, track_extraction_handler, num_workers
    )

    output_paths = sorted(output_paths)

    if verify and len(album.tracks) > 0:
        res = await verify_album(output_paths, album.tracks[0].file, num_workers)
        if not res:
            raise RuntimeError('Not bit-perfect')

    return output_paths


async def verify_album(tracks: list[Path], original: Path, num_workers) -> bool:
    result = False

    with (
        tempfile.NamedTemporaryFile(delete=True) as rhs_flac,
        tempfile.NamedTemporaryFile(delete=True) as lhs_raw,
        tempfile.NamedTemporaryFile(delete=True) as rhs_raw,
    ):
        await join_tracks(tracks, Path(rhs_flac.name))

        queue = [
            (original.resolve(), Path(lhs_raw.name).resolve()),
            (Path(rhs_flac.name).resolve(), Path(rhs_raw.name).resolve()),
        ]

        await execute_by_workers(queue, get_raw_pcm, num_workers)

        print('CMP', Path(lhs_raw.name).resolve(), Path(rhs_raw.name).resolve())
        result = await cmp_raw_pcm(
            Path(lhs_raw.name).resolve(), Path(rhs_raw.name).resolve()
        )
        print('RESULT', result)
    return result



from __future__ import annotations

from pathlib import Path
from cuetools import TrackData, AlbumData

from cuesplitter.ffmpeg import get_duration


class Track(TrackData):
    duration: float  # in seconds
    offset: float  # in seconds

    model_config = {'frozen': True}

    @classmethod
    def from_track_data(
        cls, track: TrackData, base_dir: Path, duration: float, offset: float
    ) -> Track:
        parent = track.model_dump()
        parent['file'] = (base_dir / parent['file']).resolve()

        return cls(duration=duration, offset=offset, **parent)


class Album(AlbumData):
    tracks: list[Track]  # type: ignore

    @classmethod
    async def from_album_data(cls, album: AlbumData, base_dir: Path) -> Album:
        tracks = []
        duration: float
        offset: float | None = None
        for track_cue in sorted(album.tracks, reverse=True, key=lambda x: x.track):
            offset, duration = await _get_offset_duration(
                str(base_dir), track_cue, offset
            )
            tracks.append(Track.from_track_data(track_cue, base_dir, duration, offset))

        parent = album.model_dump()
        parent['tracks'] = tracks

        return cls(**parent)


async def _get_offset_duration(
    current_dir: str, track_cue: TrackData, next_offset: float | None
) -> tuple[float, float]:
    offset = 0.0
    if track_cue.index00 is not None:
        offset = track_cue.index00.seconds
    elif track_cue.index01 is not None:
        offset = track_cue.index01.seconds
    duration = (
        next_offset - offset
        if next_offset is not None
        else await get_audiofile_duration(current_dir / track_cue.file) - offset
    )
    return offset, duration


async def get_audiofile_duration(path_to_file: Path) -> float:
    return await get_duration(path_to_file)



from pathlib import Path

import asyncio
import tempfile


async def run_cmd_raw(cmd: list[str]) -> bytes:
    sub_proccess = await asyncio.create_subprocess_exec(
        *cmd,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.DEVNULL,
    )

    result, _ = await sub_proccess.communicate()

    if sub_proccess.returncode != 0:
        raise RuntimeError()

    return result


async def run_cmd(cmd: list[str]) -> str:
    res = await run_cmd_raw(cmd)
    return res.decode().strip()


async def get_duration(audio_file: Path) -> float:
    cmd = [
        'ffprobe',  # Call ffprobe
        '-v',
        'error',  # Log level: errors
        '-show_entries',  # Specify the fields to extract
        'format=duration',  # Print only duration
        '-of',
        'default=nw=1:nokey=1',  # Without separators, etc.
        str(audio_file),
    ]

    try:
        return float(await run_cmd(cmd))
    except (RuntimeError, ValueError):
        raise RuntimeError(f'Cant proccess {audio_file} audio file duration')


async def get_bit_depth(audio_file: Path) -> float:
    cmd = [
        'ffprobe',  # Call ffprobe
        '-v',
        'error',  # Log level: errors
        '-show_entries',  # Specify the fields to extract
        'stream=bits_per_raw_sample',  # Get the bits_per_raw_sample from STREAMINFO
        '-of',
        'default=nw=1:nokey=1',  # Without separators, etc.
        str(audio_file),
    ]
    try:
        return int(await run_cmd(cmd))
    except (ValueError, RuntimeError):
        raise RuntimeError(f'Cant proccess {audio_file} audio file bit depth')


async def extract_track(
    offset: float, duration: float, input: Path, output: Path
) -> None:
    cmd = [
        'ffmpeg',  # Call ffmpeg
        '-ss',  # Start time
        str(offset),
        '-t',  # Duration
        str(duration),
        '-i',  # Input file
        str(input),
        '-c:a',
        'flac',  # Write the result into flac
        str(output),  # Output file
        '-y',  # Overwrite the output file if it already exists
    ]

    try:
        await run_cmd(cmd)
    except RuntimeError:
        raise RuntimeError(
            f'Cant extract tarck from {input} audio file. Offset: {offset}, duration: {duration}'
        )


async def join_tracks(tracks: list[Path], output: Path) -> None:
    with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.txt') as f:
        for p in tracks:
            safe_path = p.resolve().as_posix()
            f.write(f"file '{safe_path}'\n")
        filelist = f.name

    cmd = [
        'ffmpeg',  # Call ffmpeg
        '-f',
        'concat',
        '-safe',
        '0',  # Disable concat safe mode
        '-i',  # Input file
        filelist,
        '-c:a',
        'flac',  # Write the result into flac
        '-f',  # Output format
        'flac',
        str(output.resolve()),
        '-y',  # Overwrite the output file if it already exists
    ]
    try:
        await run_cmd(cmd)
    except RuntimeError:
        raise RuntimeError(f'Cant join tracks into file: {output}')

    finally:
        Path(filelist).unlink()

    print('join_tracks DONE')


async def get_raw_pcm(input: Path, output: Path) -> None:
    try:
        bit_depth = await get_bit_depth(input)
    except RuntimeError:
        bit_depth = 32

    bit_depth = bit_depth if bit_depth in [16, 24, 32] else 32

    cmd = [
        'ffmpeg',  # Call ffmpeg
        '-i',  # Input file
        str(input),
        '-f',  # Output format
        f's{bit_depth}le',
        '-acodec',  # Audio codec
        f'pcm_s{bit_depth}le',
        '-',  # Output to stdout
    ]

    try:
        pcm = await run_cmd_raw(cmd)
    except RuntimeError:
        raise RuntimeError(f'Cant get raw pcm from {input} audio file')

    with open(output, 'wb') as f:
        f.write(pcm)

    print('get_raw_pcm DONE')


async def cmp_raw_pcm(lhs: Path, rhs: Path) -> bool:
    print_file_sizes_exact(lhs, rhs)
    cmd = [
        'cmp',
        str(lhs),
        str(rhs),
    ]
    try:
        res = await run_cmd(cmd)
        print('cmp_raw_pcm DONE\n', res)
    except RuntimeError:
        return False
    return True


def print_file_sizes_exact(lhs: Path, rhs: Path) -> None:
    """Print exact file sizes in bytes."""

    def size_or_error(p: Path) -> str:
        return f'{p.stat().st_size} bytes' if p.exists() else 'NOT FOUND'

    print(f'lhs: {lhs} → {size_or_error(lhs)}')
    print(f'rhs: {rhs} → {size_or_error(rhs)}')
